#include <idc.idc>

#define vrai    1
#define faux    0

///////////////////
// operandes types
//////////////////////////////////////////////////
#define none        0         // size = 0 octets   
#define reg_field   1         // size = 1 octets  
#define reg_attr    2         // size = 2 octets  
#define immed32     3         // size = 4 octets  
#define label       4         // size = 4 octets  

///////////////////
// Attributs types
///////////////////////////
#define attr_reg32      0
#define attr_byte_ptr   1
#define attr_word_ptr   2
#define attr_dword_ptr  3
#define attr_reg8       4
#define attr_reg16      5

///////////////////
// opcode infos
///////////////////////////////
#define index_opcode        0  
#define index_name_opcode   1 
#define index_size_opcode   2  
#define index_operande1     3  
#define index_operande2     4 
#define index_special       5
//////////////////////////////////////////////////
//////////////////////////////////////////////////

static go_disam_vm (base,size)
{
    auto table_id_opcode;

    //////////////////////////////
    // Crer la table des opcodes
    ///////////////////////////////  
    table_id_opcode = CreateArray ("table_id_opcode");
    if (table_id_opcode==-1) return;

    ////////////////////////////////////////////////////////
    // Initialiser les infos sur les instructions de la VM
    ////////////////////////////////////////////////////////    
    init_opcode (table_id_opcode);

    //////////////////////////////////
    // Dsassemble le code de la VM
    ////////////////////////////////// 
    disam (base,size,table_id_opcode);
}

static disam (base,size,table_id_opcode)
{
    auto ea_scan;
    auto i;
    auto max_instr;
    auto flag;
    auto buff;

    auto string_intruction;  
    auto obj_opcode;
    auto size_opcode;
    auto operande_type_1;
    auto operande_type_2;
    auto special;
    auto operande;

    auto dword_global_id;	

    ///////////////////////////
    // Crer le buffer global
    ///////////////////////////
    dword_global_id = CreateArray ("size_global_id");
    if ( dword_global_id == -1 )
    {
        Warning("Impossible de crer le buffer global !");
        return;
    }  

    ////////////////////////////////////////////////
    // Annuler toute l'analyse sur le code de la VM
    ////////////////////////////////////////////////   
    for (i=0;i<size;i++) MakeUnkn (base+i,0);
    ///////////////////////////////////////////// 

    ea_scan = base;
    max_instr = GetLastIndex (AR_LONG,table_id_opcode)+1;

    while (ea_scan<base+size)
    {    
        flag=0;
        for (i=0;i<max_instr;i++)
        {
            obj_opcode = GetArrayElement (AR_LONG,table_id_opcode,i);
            if (Byte(ea_scan)!=GetArrayElement (AR_LONG,obj_opcode,index_opcode)) continue;

            flag=1; 

            size_opcode       = GetArrayElement (AR_LONG,obj_opcode,index_size_opcode);
            operande_type_1   = GetArrayElement (AR_LONG,obj_opcode,index_operande1);
            operande_type_2   = GetArrayElement (AR_LONG,obj_opcode,index_operande2);
            special           = GetArrayElement (AR_LONG,obj_opcode,index_special);
            string_intruction = GetArrayElement (AR_STR,obj_opcode,index_name_opcode);      

            if (operande_type_1!=0)
            {
                SetArrayLong (dword_global_id,0,0);
                get_operande (ea_scan,obj_opcode,1,dword_global_id);
                operande = GetArrayElement(AR_LONG,dword_global_id,0);

                SetArrayString (dword_global_id,0,0);	
                if (operande_type_2==0 && special!=0)
                {	         
                    operd2string (ea_scan,operande_type_1,operande, special,dword_global_id);
                }
                else
                {
                    operd2string (ea_scan,operande_type_1,operande, 0,dword_global_id);                  
                }

                buff = GetArrayElement(AR_STR,dword_global_id,0);

                string_intruction = string_intruction + buff;

                if (operande_type_2!=0)
                {         
                    SetArrayLong (dword_global_id,0,0);
                    get_operande (ea_scan,obj_opcode,2,dword_global_id);
                    operande = GetArrayElement(AR_LONG,dword_global_id,0);          

                    SetArrayString (dword_global_id,0,0); 
                    operd2string (ea_scan,operande_type_2, operande, special,dword_global_id);                  
                    buff = GetArrayElement(AR_STR,dword_global_id,0); 

                    string_intruction = string_intruction + ", "+ buff;               
                }         
            }

            MakeByte (ea_scan);
            MakeArray (ea_scan, size_opcode);
            SetManualInsn (ea_scan,string_intruction);

            ea_scan = ea_scan + size_opcode;
            break;
        }  
        if (flag==0)
        {
            Warning ("Erreur : opcode inconnu ! -> 0x"+ ltoa (Byte(ea_scan)));
            SetManualInsn (ea_scan,"bad opcode");
            ea_scan=base+size;
            break;
        }
    }

    /////////////
    // Nettoyer
    /////////////		
    DeleteArray (dword_global_id);    
}


static get_operande (ea_scan,obj_opcode,operande_number,dword_global_id)
{
    auto resultat;
    auto index;
    auto operande_type;
    auto op_type_1;
    auto op_type_2;

    operande_type=0;
    resultat=0;
    op_type_1 = GetArrayElement (AR_LONG,obj_opcode,index_operande1);
    op_type_2 = GetArrayElement (AR_LONG,obj_opcode,index_operande2);

    index=1;                    // saute l'opcode (1 octet)

    if (operande_number==1) operande_type  = op_type_1;
    else
        if (operande_number==2)
        {                           // Saute la premiere operande
            if (op_type_1==reg_field) index++;
            else
                if (op_type_1==reg_attr) index=index+2; 
                else
                    if (op_type_1==immed32 || op_type_1==label) index=index+4;
            operande_type = op_type_2;
        }

        if (operande_type==reg_field) resultat=Byte(ea_scan+index); 
        else
            if (operande_type==reg_attr) resultat=Word(ea_scan+index);
            else
                if (operande_type==immed32 || operande_type==label) resultat=Dword(ea_scan+index); 

        SetArrayLong (dword_global_id,0,resultat);
}

static operd2string (ea_scan, operande_type, operande, special,dword_global_id)
{
    auto reg;
    auto attr;
    auto chaine;
    auto buff;
    auto @label;

    chaine=0;
    buff=0;
    @label=0;

    while (vrai)
    {
        /////////////////////
        // Type : reg_field
        /////////////////////
        if (operande_type==reg_field)
        {
            chaine="reg32_"+ ltoa (operande,16); 
            SetArrayString (dword_global_id,chaine,0);    
            break;
        }
        /////////////////////
        // Type : reg/attr
        /////////////////////   

        if (operande_type==reg_attr)
        {
            reg = (operande>>8)& 0xFF;
            attr = operande & 0xFF;
            if (attr==attr_reg32) chaine="reg32_"+ltoa(reg,16);
            else 
                if (attr==attr_byte_ptr) chaine="byte ptr ["+"reg32_"+ltoa(reg,16)+"]";
                else
                    if (attr==attr_word_ptr) chaine="word ptr ["+"reg32_"+ltoa(reg,16)+"]";
                    else
                        if (attr==attr_dword_ptr) chaine="dword ptr ["+"reg32_"+ltoa(reg,16)+"]";
                        else
                            if (attr==attr_reg8) chaine="reg8_"+ltoa(reg,16);
                            else
                                if (attr==attr_reg16) chaine="reg16_"+ltoa(reg,16); 
            break;
        } 
        /////////////////////
        // Type : immed32
        /////////////////////
        if (operande_type==immed32)
        {
            buff=operande+special;              
            @label=NameEx (0,buff);
            if (@label!="") chaine = @label;
            else chaine="0"+ltoa (buff,16)+"h";

            break;
        } 
        /////////////////////
        // Type : label
        /////////////////////
        if (operande_type==label)
        {
            buff=ea_scan+operande+special;
            chaine="VM_loc_"+ltoa (buff,16);
            MakeName (buff, chaine);
            break;
        }    
        break;
    }
    SetArrayString (dword_global_id,0,chaine);
}

static init_opcode (table_id_opcode)
{
    auto id_opcode;
    auto index;


    index = 0;

    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x56);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_mov       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 2);                // id_operande1
    SetArrayLong    (id_opcode, 4, 2);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x89);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_push      ");  // Name
    SetArrayLong    (id_opcode, 2, 2);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x33);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_pop       ");  // Name
    SetArrayLong    (id_opcode, 2, 2);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xE4);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_nop       ");  // Name
    SetArrayLong    (id_opcode, 2, 1);                // size
    SetArrayLong    (id_opcode, 3, 0);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xF4);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_exit      ");  // Name
    SetArrayLong    (id_opcode, 2, 1);                // size
    SetArrayLong    (id_opcode, 3, 0);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xA0);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_xor       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 2);                // id_operande1
    SetArrayLong    (id_opcode, 4, 2);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x65);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_inc       ");  // Name
    SetArrayLong    (id_opcode, 2, 2);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x99);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_dec       ");  // Name
    SetArrayLong    (id_opcode, 2, 2);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xE8);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_add       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 2);                // id_operande1
    SetArrayLong    (id_opcode, 4, 2);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xF2);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sub       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 2);                // id_operande1
    SetArrayLong    (id_opcode, 4, 2);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xCE);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_cmp       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 2);                // id_operande1
    SetArrayLong    (id_opcode, 4, 2);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xCC);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_jmp       ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 4);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0);                // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x78);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_mul    ");  // Name
    SetArrayLong    (id_opcode, 2, 6);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 3);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0xEA9DBA9E);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xD3);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_add    ");  // Name
    SetArrayLong    (id_opcode, 2, 6);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 3);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0xCE072A5D);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x96);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_jz     ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 4);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0x40952466);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x97);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_jnz    ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 4);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0x21AE13F7);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x57);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_jc     ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 4);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0x74F364CA);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x17);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_mov    ");  // Name
    SetArrayLong    (id_opcode, 2, 6);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 3);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0x77E72856);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0x67);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_jnc    ");  // Name
    SetArrayLong    (id_opcode, 2, 5);                // size
    SetArrayLong    (id_opcode, 3, 4);                // id_operande1
    SetArrayLong    (id_opcode, 4, 0);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0x37B91026);       // special

    // -------------------------------------------------------------
    index++;
    id_opcode = CreateArray ("opcode_id_"+ ltoa (index,10));
    SetArrayLong (table_id_opcode, index, id_opcode);

    SetArrayLong    (id_opcode, 0, 0xDF);             // Opcode
    SetArrayString  (id_opcode, 1, "vm_sp_sub    ");  // Name
    SetArrayLong    (id_opcode, 2, 6);                // size
    SetArrayLong    (id_opcode, 3, 1);                // id_operande1
    SetArrayLong    (id_opcode, 4, 3);                // id_operande2  
    SetArrayLong    (id_opcode, 5, 0xD7A156E2);       // special    
}

/*
Opcode    Name        Size  ID_op1  ID_op2  spcial       double_spcial      
0x56      vm_mov      5       2       2       0                       
0x89      vm_push     2       1       0       0                       
0x33      vm_pop      2       1       0       0                       
0xE4      vm_nop      1       0       0       0                       
0xF4      vm_exit     1       0       0       0                       
0xA0      vm_xor      5       2       2       0                       
0x65      vm_inc      2       1       0       0                       
0x99      vm_dec      2       1       0       0                       
0xE8      vm_add      5       2       2       0                       
0xF2      vm_sub      5       2       2       0                       
0xCE      vm_cmp      5       2       2       0                       
0xCC      vm_jmp      5       4       0       0                       
0x78      vm_sp_mul   6       1       3       0xEA9DBA9E  (-1)*0x15624562
0xD3      vm_sp_add   6       1       3       0xCE072A5D  (-1)*0x31F8D5A3
0x96      vm_sp_jz    5       4       0       0x40952466  (-1)*0xBF6ADB9A
0x97      vm_sp_jnz   5       4       0       0x21AE13F7      
0x57      vm_sp_jc    5       4       0       0x74F364CA      
0x17      vm_sp_mov   6       1       3       0x77E72856  
0x67      vm_sp_jnc   5       4       0       0x37B91026  (-1)*0xC846EFDA      
0xDF      vm_sp_sub   6       1       3       0xD7A156E2  
*/
